import re
import json
import logging
from pydash import filter_
from .selections import build_cypher_selection
from .utils import (is_mutation, is_add_relationship_mutation, type_identifiers, low_first_letter, cypher_directive,
                    mutation_meta_directive, extract_query_result, extract_selections)

logger = logging.getLogger('neo4j_graphql_py')
logger.setLevel(logging.DEBUG)
ch = logging.StreamHandler()
formatter = logging.Formatter('%(levelname)s:     %(message)s')
ch.setFormatter(formatter)
logger.addHandler(ch)


def neo4j_graphql(obj, context, resolve_info, debug=False, **kwargs):
    if is_mutation(resolve_info):
        query = cypher_mutation(context, resolve_info, **kwargs)
        if not is_add_relationship_mutation(resolve_info):
            kwargs = {'params': kwargs}
    else:
        query = cypher_query(context, resolve_info, **kwargs)
    if debug:
        logger.info(query)
        logger.info(kwargs)

    with context.get('driver').session() as session:
        data = session.run(query, **kwargs)
        data = extract_query_result(data, resolve_info.return_type)
        return data


def cypher_query(context, resolve_info, first=-1, offset=0, _id=None, **kwargs):
    types_ident = type_identifiers(resolve_info.return_type)
    type_name = types_ident.get('type_name')
    variable_name = types_ident.get('variable_name')
    schema_type = resolve_info.schema.get_type(type_name)

    filtered_field_nodes = filter_(resolve_info.field_nodes, lambda n: n.name.value == resolve_info.field_name)

    # FIXME: how to handle multiple field_node matches
    selections = extract_selections(filtered_field_nodes[0].selection_set.selections, resolve_info.fragments)

    # if len(selections) == 0:
    #     # FIXME: why aren't the selections found in the filteredFieldNode?
    #     selections = extract_selections(resolve_info.operation.selection_set.selections, resolve_info.fragments)

    # FIXME: support IN for multiple values -> WHERE
    arg_string = re.sub(r"\"([^(\")]+)\":", "\\1:", json.dumps(kwargs))

    id_where_predicate = f'WHERE ID({variable_name})={_id} ' if _id is not None else ''
    outer_skip_limit = f'SKIP {offset}{" LIMIT " + str(first) if first > -1 else ""}'

    cyp_dir = cypher_directive(resolve_info.schema.query_type, resolve_info.field_name)
    if cyp_dir:
        custom_cypher = cyp_dir.get('statement')
        query = (f'WITH apoc.cypher.runFirstColumn("{custom_cypher}", {arg_string}, true) AS x '
                 f'UNWIND x AS {variable_name} RETURN {variable_name} '
                 f'{{{build_cypher_selection("", selections, variable_name, schema_type, resolve_info)}}} '
                 f'AS {variable_name} {outer_skip_limit}')
    else:
        # No @cypher directive on QueryType
        query = f'MATCH ({variable_name}:{type_name} {arg_string}) {id_where_predicate}'
        query += (f'RETURN {variable_name} '
                  f'{{{build_cypher_selection("", selections, variable_name, schema_type, resolve_info)}}}'
                  f' AS {variable_name} {outer_skip_limit}')

    return query


def cypher_mutation(context, resolve_info, first=-1, offset=0, _id=None, **kwargs):
    # FIXME: lots of duplication here with cypherQuery, extract into util module
    types_ident = type_identifiers(resolve_info.return_type)
    type_name = types_ident.get('type_name')
    variable_name = types_ident.get('variable_name')
    schema_type = resolve_info.schema.get_type(type_name)

    filtered_field_nodes = filter_(resolve_info.field_nodes, lambda n: n.name.value == resolve_info.field_name)

    # FIXME: how to handle multiple field_node matches
    selections = extract_selections(filtered_field_nodes[0].selection_set.selections, resolve_info.fragments)

    # FIXME: support IN for multiple values -> WHERE
    arg_string = re.sub(r"\"([^(\")]+)\":", "\\1:", json.dumps(kwargs))

    id_where_predicate = f'WHERE ID({variable_name})={_id} ' if _id is not None else ''
    outer_skip_limit = f'SKIP {offset}{" LIMIT " + str(first) if first > -1 else ""}'

    cyp_dir = cypher_directive(resolve_info.schema.mutation_type, resolve_info.field_name)
    if cyp_dir:
        custom_cypher = cyp_dir.get('statement')
        query = (f'CALL apoc.cypher.doIt("{custom_cypher}", {arg_string}) YIELD value '
                 f'WITH apoc.map.values(value, [keys(value)[0]])[0] AS {variable_name} '
                 f'RETURN {variable_name} {{{build_cypher_selection("", selections, variable_name, schema_type, resolve_info)}}} '
                 f'AS {variable_name} {outer_skip_limit}')
    # No @cypher directive on MutationType
    elif resolve_info.field_name.startswith('create') or resolve_info.field_name.startswith('Create'):
        # Create node
        # TODO: handle for create relationship
        # TODO: update / delete
        # TODO: augment schema
        query = (f'CREATE ({variable_name}:{type_name}) SET {variable_name} = $params RETURN {variable_name} '
                 f'{{{build_cypher_selection("", selections, variable_name, schema_type, resolve_info)}}} '
                 f'AS {variable_name}')
    elif resolve_info.field_name.startswith('add') or resolve_info.field_name.startswith('Add'):
        mutation_meta = mutation_meta_directive(resolve_info.schema.mutation_type, resolve_info.field_name)
        relation_name = mutation_meta.get('relationship')
        from_type = mutation_meta.get('from')
        from_var = low_first_letter(from_type)
        to_type = mutation_meta.get('to')
        to_var = low_first_letter(to_type)
        from_param = resolve_info.schema.mutation_type.fields[resolve_info.field_name].ast_node.arguments[0].name.value
        to_param = resolve_info.schema.mutation_type.fields[resolve_info.field_name].ast_node.arguments[1].name.value
        query = (f'MATCH ({from_var}:{from_type} {{{from_param}: ${from_param}}}) '
                 f'MATCH ({to_var}:{to_type} {{{to_param}: ${to_param}}}) '
                 f'CREATE ({from_var})-[:{relation_name}]->({to_var}) '
                 f'RETURN {from_var} '
                 f'{{{build_cypher_selection("", selections, variable_name, schema_type, resolve_info)}}} '
                 f'AS {from_var}')
    else:
        raise Exception('Mutation does not follow naming conventions')
    return query


def augment_schema(schema):
    from .augment_schema import add_mutations_to_schema
    mutation_schema = add_mutations_to_schema(schema)
    return mutation_schema



from .main import neo4j_graphql, cypher_query, cypher_mutation, augment_schema
from .utils import make_executable_schema

__all__ = [
    "neo4j_graphql",
    "cypher_query",
    "cypher_mutation",
    "augment_schema",
    "make_executable_schema",
]



from .utils import (cypher_directive_args, is_graphql_scalar_type, is_array_type, inner_type, cypher_directive,
                    relation_directive, inner_filter_params, compute_skip_limit)


def build_cypher_selection(initial, selections, variable_name, schema_type, resolve_info):
    if len(selections) == 0:
        return initial
    head_selection, *tail_selections = selections

    tail_params = {
        'selections': tail_selections,
        'variable_name': variable_name,
        'schema_type': schema_type,
        'resolve_info': resolve_info
    }

    field_name = head_selection.name.value
    comma_if_tail = ',' if len(tail_selections) > 0 else ''
    # Schema meta fields(__schema, __typename, etc)
    if not schema_type.fields.get(field_name):
        return build_cypher_selection(initial[1:initial.rfind(',')] if len(tail_selections) == 0 else initial,
                                      **tail_params)

    field_type = schema_type.fields[field_name].type

    inner_schema_type = inner_type(field_type)  # for target "field_type" aka label

    custom_cypher = cypher_directive(schema_type, field_name).get('statement')

    # Database meta fields(_id)
    if field_name == '_id':
        return build_cypher_selection(f'{initial}{field_name}: ID({variable_name}){comma_if_tail}', **tail_params)
    # Main control flow
    if is_graphql_scalar_type(inner_schema_type):
        if custom_cypher:
            return build_cypher_selection((f'{initial}{field_name}: apoc.cypher.runFirstColumn("{custom_cypher}", '
                                           f'{cypher_directive_args(variable_name, head_selection, schema_type, resolve_info)}, false)'
                                           f'{comma_if_tail}'), **tail_params)

        # graphql scalar type, no custom cypher statement
        return build_cypher_selection(f'{initial} .{field_name} {comma_if_tail}', **tail_params)

    # We have a graphql object type
    nested_variable = variable_name + '_' + field_name
    skip_limit = compute_skip_limit(head_selection, resolve_info.variable_values)
    nested_params = {
        'initial': '',
        'selections': head_selection.selection_set.selections,
        'variable_name': nested_variable,
        'schema_type': inner_schema_type,
        'resolve_info': resolve_info
    }
    if custom_cypher:
        # similar: [ x IN apoc.cypher.runFirstColumn("WITH {this} AS this MATCH (this)--(:Genre)--(o:Movie)
        # RETURN o", {this: movie}, true) |x {.title}][1..2])

        field_is_list = not not getattr(field_type, 'of_type', None)

        return build_cypher_selection(
            (f'{initial}{field_name}: {"" if field_is_list else "head("}'
             f'[ {nested_variable} IN apoc.cypher.runFirstColumn("{custom_cypher}", '
             f'{cypher_directive_args(variable_name, head_selection, schema_type, resolve_info)}, true) | {nested_variable} '
             f'{{{build_cypher_selection(**nested_params)}}}]'
             f'{"" if field_is_list else ")"}{skip_limit} {comma_if_tail}'), **tail_params)

    # graphql object type, no custom cypher

    rel = relation_directive(schema_type, field_name)
    rel_type = rel.get('name')
    rel_direction = rel.get('direction')
    subquery_args = inner_filter_params(head_selection)

    return build_cypher_selection(
        (f"{initial}{field_name}: {'head(' if not is_array_type(field_type) else ''}"
         f"[({variable_name}){'<' if rel_direction in ['in', 'IN'] else ''}"
         f"-[:{rel_type}]-{'>' if rel_direction in ['out', 'OUT'] else ''}"
         f"({nested_variable}:{inner_schema_type.name} {subquery_args}) | {nested_variable} "
         f"{{{build_cypher_selection(**nested_params)}}}]"
         f"{')' if not is_array_type(field_type) else ''}{skip_limit} {comma_if_tail}"), **tail_params)



from .utils import inner_type, make_executable_schema
from .main import neo4j_graphql
from graphql import print_schema
from pydash import filter_, reduce_


def add_mutations_to_schema(schema):
    types = types_to_augment(schema)

    # FIXME: don't use printSchema (custom directives are lost), instead use extend schema
    # FIXME: type extensions are lost
    mutation_schema_sdl = print_schema(schema)

    # TODO: compose augment funcs
    # let mutationSchemaSDLWithTypes = augmentTypes(types, schema, mutationSchemaSDL);

    mutation_schema_sdl_with_types_and_mutations = augment_mutations(types, schema, mutation_schema_sdl)

    def resolve_neo4j(obj, info, **kwargs):
        return neo4j_graphql(obj, info.context, info, **kwargs)

    # console.log(mutationSchemaSDLWithTypesAndMutations);
    def f(acc, t):
        # FIXME: inspect actual mutations, not construct mutation names here
        acc['Mutation'][f'Create{t}'] = resolve_neo4j
        for field_type in types:
            for rel_mutation in add_relationship_mutations(schema.type_map[field_type], True):
                acc['Mutation'][rel_mutation] = resolve_neo4j
        return acc

    resolvers = reduce_(types, f, {'Query': {}, 'Mutation': {}})

    # delegate query resolvers to original schema
    def f2(acc, t):
        acc['Query'][t] = resolve_neo4j
        return acc

    resolvers = reduce_(list(schema.query_type.fields.keys()), f2, resolvers)

    mutation_schema = make_executable_schema(mutation_schema_sdl_with_types_and_mutations, resolvers)

    final_schema = mutation_schema
    return final_schema


def types_to_augment(schema):
    """
     * Given a GraphQLSchema return an array of the type names,
     * excluding Query and Mutation types
     * @param {GraphQLSchema} schema
     * @returns {string[]}
    """
    # TODO: check for @ignore and @model directives
    return filter_(list(schema.type_map.keys()),
                   lambda t: False if schema.type_map[t].ast_node is None
                   else schema.type_map[
                            t].ast_node.kind == 'object_type_definition' and t != 'Query' and t != 'Mutation')


def augment_types(types, schema, sdl):
    """
    #  * Generate type extensions for each type:
    #  *   - add _id field
    #  * @param {string[]} types
    #  * @param schema
    #  * @param {string} sdl
    #  * @returns {string} SDL type extensions
    """
    return reduce_(types,
                   lambda acc, t: acc + '' if t in ['Mutation', 'Query']
                   else acc + f'extend type {t} {{ _id:ID }}', sdl)


def augment_mutations(types, schema, sdl):
    # FIXME: requires placeholder Query type
    return (sdl +
            f'''
            extend schema {{
              mutation: Mutation
            }}
        
        
            type Mutation {{
                {reduce_(types, lambda acc, t: acc + f'{create_mutation(schema.type_map[t])} '
                                                     f'{add_relationship_mutations(schema.type_map[t])} ', '')}
            }}
            '''
            )


def create_mutation(field_type):
    return f'Create{field_type.name}({param_signature(field_type)}): {field_type.name}'


def add_relationship_mutations(field_type, names_only=False):
    mutations = ''
    mutation_names = []

    def f1(field):
        i = 0
        while i < len(field.ast_node.directives):
            return field.ast_node.directives[i].name.value == 'relation'

    relationship_fields = filter_(field_type.fields, f1)

    for field in relationship_fields:
        relation_directive = (
            filter_(field.ast_node.directives, lambda d: d.name.value == 'relation')[0])
        rel_type = filter_(relation_directive.arguments, lambda a: a.name.value == 'name')[0]
        rel_direction = filter_(relation_directive.arguments, lambda a: a.name.value == 'direction')[0]

        if rel_direction.value.value in ['out', 'OUT']:
            from_type = field_type
            to_type = inner_type(field.type)
        else:
            from_type = inner_type(field.type)
            to_type = field_type
        from_pk = primary_key(from_type)
        to_pk = primary_key(to_type)

        # FIXME: could add relationship properties here
        mutations += (f'Add{from_type.name}{to_type.name}'
                      f'({from_pk.ast_node.name.value}: {inner_type(from_pk.type).name}!, '
                      f'{to_pk.ast_node.name.value}: {inner_type(to_pk.type).name}!): '
                      f'{from_type.name} @MutationMeta(relationship: "{rel_type.value.value}", from: "{from_type.name}", to: "{to_type.name}")')
        mutation_names.append(f'Add{from_type.name}{to_type.name}')
    if names_only:
        return mutation_names
    else:
        return mutations


def primary_key(field_type):
    """
     * Returns the field to be treated as the "primary key" for this type
     * Primary key is determined as the first of:
     *   - non-null ID field
     *   - ID field
     *   - first String field
     *   - first field
     *
     * @param {object_type_definition} type
     * @returns {FieldDefinition} primary key field
    """
    # Find the primary key for the type
    # first field with a required ID
    # if no required ID type then first required type
    pk = first_non_null_and_id_field(field_type)
    if not pk:
        pk = first_id_field(field_type)
    if not pk:
        pk = first_non_null_field(field_type)
    if not pk:
        pk = first_field(field_type)

    return pk


def param_signature(field_type):
    def fun(acc, f):
        if f == '_id' or (getattr(inner_type(field_type.fields[f].type), 'ast_node', None) is not None and inner_type(
                field_type.fields[f].type).ast_node.kind == 'object_type_definition'):
            # TODO: exclude @cypher fields
            # TODO: exclude object types?
            return acc + ''
        else:
            return acc + f' {f}: {inner_type(field_type.fields[f].type).name}, '

    return reduce_(list(field_type.fields.keys()), fun, '')


def first_non_null_and_id_field(field_type):
    fields = filter_(list(field_type.fields.keys()),
                     lambda t: type(field_type.fields[t]).__name__ == 'GraphQLNonNull'
                               and field_type.fields[t].type.name == 'ID')
    if len(fields) > 0:
        return field_type.fields[fields[0]]
    else:
        return None


def first_id_field(field_type):
    fields = filter_(list(field_type.fields.keys()), lambda t: inner_type(field_type.fields[t].type).name == 'ID')
    if len(fields) > 0:
        return field_type.fields[fields[0]]
    else:
        return None


def first_non_null_field(field_type):
    fields = filter_(list(field_type.fields.keys()), lambda t: type(field_type.fields[t]).__name__ == 'GraphQLNonNull')
    if len(fields) > 0:
        return field_type.fields[fields[0]]
    else:
        return None


def first_field(field_type):
    return field_type.fields[list(field_type.fields.keys())[0]]



import re
import json
import logging
from typing import Any
from pydash import find, reduce_
from graphql import GraphQLResolveInfo, GraphQLScalarType, parse, build_ast_schema

logger = logging.getLogger('neo4j_graphql_py')


def make_executable_schema(schema_definition, resolvers):
    ast = parse(schema_definition)
    schema = build_ast_schema(ast, assume_valid=True)

    for type_name in resolvers:
        field_type = schema.get_type(type_name)

        for field_name in resolvers[type_name]:
            if field_type is GraphQLScalarType:
                field_type.fields[field_name].resolve = resolvers[type_name][field_name]
                continue

            field = field_type.fields[field_name]
            field.resolve = resolvers[type_name][field_name]

        if not field_type.fields:
            continue

        for remaining in field_type.fields:
            if field_type.fields[remaining].resolve is None:
                field_type.fields[remaining].resolve = default_resolver

    return schema


def default_resolver(source: Any, info: GraphQLResolveInfo, **args: Any) -> Any:
    field_name = info.field_name
    value = (
        source.get(field_name)
        if isinstance(source, dict)
        else getattr(source, field_name, None)
    )
    if callable(value):
        return value(info, **args)
    return value


def parse_args(args, variable_values):
    if args is None or len(args) == 0:
        return {}

    return {arg.name.value: (
        int(arg.value.value) if arg.value.kind == 'int_value'
        else float(arg.value.value) if arg.value.kind == 'float_value'
        else variable_values[arg.name.value] if arg.value.kind == 'variable'
        else arg.value.value)
        for arg in args}


def get_default_arguments(field_name, schema_type):
    # get default arguments for this field from schema
    args = schema_type.fields[field_name].args
    return {arg_name: arg.default_value for arg_name, arg in args.items()}


def cypher_directive_args(variable, head_selection, schema_type, resolve_info):
    default_args = get_default_arguments(head_selection.name.value, schema_type)
    schema_args = {}
    query_args = parse_args(head_selection.arguments, resolve_info.variable_values)
    default_args.update(query_args)
    args = re.sub(r"\"([^(\")]+)\":", "\\1:", json.dumps(default_args))
    return f'{{this: {variable}{args[1:]}' if args == "{}" else f'{{this: {variable}, {args[1:]}'


def is_mutation(resolve_info):
    return resolve_info.operation.operation == 'mutation'


def is_add_relationship_mutation(resolve_info):
    return (is_mutation(resolve_info)
            and
            (resolve_info.field_name.startswith('add')
             or resolve_info.field_name.startswith('Add'))
            )


def type_identifiers(return_type):
    type_name = str(inner_type(return_type))
    return {'variable_name': low_first_letter(type_name),
            'type_name': type_name}


def is_graphql_scalar_type(field_type):
    return type(field_type).__name__ == 'GraphQLScalarType' or type(field_type).__name__ == 'GraphQLEnumType'


def is_array_type(field_type):
    return str(field_type).startswith('[')


def low_first_letter(word):
    return word[0].lower() + word[1:]


def inner_type(field_type):
    return inner_type(field_type.of_type) if getattr(field_type, 'of_type', None) else field_type


def directive_with_args(directive_name, *args):
    def fun(schema_type, field_name):
        def field_directive(schema_type, field_name, directive_name):
            return find(schema_type.fields[field_name].ast_node.directives, lambda d: d.name.value == directive_name)

        def directive_argument(directive, name):
            return find(directive.arguments, lambda a: a.name.value == name).value.value

        directive = field_directive(schema_type, field_name, directive_name)
        ret = {}
        if directive:
            ret.update({key: directive_argument(directive, key) for key in args})
        return ret

    return fun


cypher_directive = directive_with_args('cypher', 'statement')
relation_directive = directive_with_args('relation', 'name', 'direction')
mutation_meta_directive = directive_with_args('MutationMeta', 'relationship', 'from', 'to')


def inner_filter_params(selections):
    query_params = {}
    if len(selections.arguments) > 0:
        query_params = {arg.name.value: arg.value.value for arg in selections.arguments if
                        arg.name.value not in ['first', 'offset']}
    # FIXME: support IN for multiple values -> WHERE
    query_params = re.sub(r"\"([^(\")]+)\":", "\\1:", json.dumps(query_params))

    return query_params


def argument_value(selection, name, variable_values):
    arg = find(selection.arguments, lambda argument: argument.name.value == name)
    return (
        None if arg is None
        else variable_values[name] if getattr(arg.value, 'value', None) is None
                                      and name in variable_values and arg.value.kind == 'variable'
        else arg.value.value
    )


def extract_query_result(records, return_type):
    type_ident = type_identifiers(return_type)
    variable_name = type_ident.get('variable_name')
    result = [record.get(variable_name) for record in records.data()]
    return result if is_array_type(return_type) else result[0] if len(result) > 0 else None


def compute_skip_limit(selection, variable_values):
    first = argument_value(selection, "first", variable_values)
    offset = argument_value(selection, "offset", variable_values)
    if first is None and offset is None:
        return ""
    if offset is None:
        return f'[..{first}]'
    if first is None:
        return f'[{offset}..]'
    return f'[{offset}..{int(offset) + int(first)}]'


def extract_selections(selections, fragments):
    # extract any fragment selection sets into a single array of selections
    return reduce_(selections,
                   lambda acc, curr:
                   [*acc, *fragments[curr.name.value].selection_set.selections] if curr.kind == 'fragment_spread'
                   else [*acc, curr],
                   [])
